EXPORTS

; Make runtime functions visible
clGetPlatformIDs
clGetPlatformInfo
clGetDeviceIDs
clGetDeviceInfo
clCreateContext
clCreateContextFromType
clRetainContext
clReleaseContext
clGetContextInfo
clCreateCommandQueue
clRetainCommandQueue
clReleaseCommandQueue
clGetCommandQueueInfo
clSetCommandQueueProperty
clCreateBuffer
clCreateImage2D
clCreateImage3D
clRetainMemObject
clReleaseMemObject
clGetSupportedImageFormats
clGetMemObjectInfo
clGetImageInfo
clCreateSampler
clRetainSampler
clReleaseSampler
clGetSamplerInfo
clCreateProgramWithSource
clCreateProgramWithBinary
clRetainProgram
clReleaseProgram
clBuildProgram
clUnloadCompiler
clGetProgramInfo
clGetProgramBuildInfo
clCreateKernel
clCreateKernelsInProgram
clRetainKernel
clReleaseKernel
clSetKernelArg
clGetKernelInfo
clGetKernelWorkGroupInfo
clWaitForEvents
clGetEventInfo
clRetainEvent
clReleaseEvent
clGetEventProfilingInfo
clFlush
clFinish
clEnqueueReadBuffer
clEnqueueWriteBuffer
clEnqueueCopyBuffer
clEnqueueReadImage
clEnqueueWriteImage
clEnqueueCopyImage
clEnqueueCopyImageToBuffer
clEnqueueCopyBufferToImage
clEnqueueMapBuffer
clEnqueueMapImage
clEnqueueUnmapMemObject
clEnqueueNDRangeKernel
clEnqueueTask
clEnqueueNativeKernel
clEnqueueMarker
clEnqueueWaitForEvents
clEnqueueBarrier
clGetExtensionFunctionAddress
clCreateFromGLBuffer
clCreateFromGLTexture2D
clCreateFromGLTexture3D
clCreateFromGLRenderbuffer
clGetGLObjectInfo
clGetGLTextureInfo
clEnqueueAcquireGLObjects
clEnqueueReleaseGLObjects
clGetGLContextInfoKHR
clGetDeviceIDsFromD3D10KHR
clCreateFromD3D10BufferKHR
clCreateFromD3D10Texture2DKHR
clCreateFromD3D10Texture3DKHR
clEnqueueAcquireD3D10ObjectsKHR
clEnqueueReleaseD3D10ObjectsKHR
clSetEventCallback
clCreateSubBuffer
clSetMemObjectDestructorCallback
clCreateUserEvent
clSetUserEventStatus
clEnqueueReadBufferRect
clEnqueueWriteBufferRect
clEnqueueCopyBufferRect
clCreateEventFromGLsyncKHR
clCreateSubDevices
clRetainDevice
clReleaseDevice
clCreateImage
clCreateProgramWithBuiltInKernels
clCompileProgram
clLinkProgram
clUnloadPlatformCompiler
clGetKernelArgInfo
clEnqueueFillBuffer
clEnqueueFillImage
clEnqueueMigrateMemObjects
clEnqueueMarkerWithWaitList
clEnqueueBarrierWithWaitList
clGetExtensionFunctionAddressForPlatform
clCreateFromGLTexture
clGetDeviceIDsFromD3D11KHR
clCreateFromD3D11BufferKHR
clCreateFromD3D11Texture2DKHR
clCreateFromD3D11Texture3DKHR
clCreateFromDX9MediaSurfaceKHR
clEnqueueAcquireD3D11ObjectsKHR
clEnqueueReleaseD3D11ObjectsKHR
clGetDeviceIDsFromDX9MediaAdapterKHR
clEnqueueAcquireDX9MediaSurfacesKHR
clEnqueueReleaseDX9MediaSurfacesKHR

initOclgrind